
#include "harktool.h"


/*-----------------------------------------------------------------------------
*  FFT 計算       (float complex --> float complex)
*
*  COMPLEX_FLOAT*  a  // 入力列
*  COMPLEX_FLOAT*  A  // 出力・計算用配列（in-place計算）
*  int fftsize        // FFT長
*  COMPLEX_FLOAT*  w  // 作業用配列（NULLの場合は関数内で確保する）
-----------------------------------------------------------------------------*/
void harktool_FFTImpl_fFFT(COMPLEX_FLOAT *a, COMPLEX_FLOAT *A, int fftsize)
{

    int m, mh, i, j, k, irev, prev_irev;
    COMPLEX_FLOAT y;
    const int n = fftsize;



    // 初期化
    if (a != A) {
        memcpy((void*) A, (void*) a, n * sizeof(COMPLEX_FLOAT));
    }
    COMPLEX_FLOAT *w = (COMPLEX_FLOAT *)malloc(n * sizeof(COMPLEX_FLOAT));
    if (w == NULL) {
        harkio_Log_printf(harkio_Log_Error, "malloc error");
        exit(EXIT_FAILURE);
    }

    for (i = 0; i < n; i++) {
        w[i] = cos(two_pi * i / n) - sin(two_pi * i / n) * _Complex_I;
    }

    // ビット反転並べ替え
    i = 0;
    for (j = 1; j < n - 1; j++) {
        for (k = n >> 1; k > (i ^= k); k >>= 1)
            ;
        if (j < i) {
            harktool_swapFloatComplex(&A[i], &A[j]);
        }
    }

    // 主要計算部
    for (mh = 1; (m = mh << 1) <= n; mh = m) {
        irev = 0;
        for (i = 0; i < n; i += m) {
            prev_irev = irev;
            // w[prev_irev] = cos(two_pi * irev / n) - sin(two_pi * irev / n) * _Complex_I;
            for (k = n >> 2; k > (irev ^= k); k >>= 1)
                ;
            for (j = i; j < mh + i; j++) {
                k = j + mh;
                y = A[j] - A[k];
                A[j] = A[j] + A[k];
                A[k] = w[prev_irev] * y;
            }
        }
    }

    // 終了処理
    free(w);
    w = NULL;
}

/*-----------------------------------------------------------------------------
*  FFT 逆計算       (float complex <-- float complex)
*
*  COMPLEX_FLOAT*  A  // 入力列
*  COMPLEX_FLOAT*  a  // 出力・計算用配列（in-place計算）
*  int fftsize        // FFT長
*  COMPLEX_FLOAT*  w  // 作業用配列（NULLの場合は関数内で確保する）
-----------------------------------------------------------------------------*/
void harktool_FFTImpl_fFFTInv(COMPLEX_FLOAT *A, COMPLEX_FLOAT *a, int fftsize)
{

    int m, mh, i, j, k, irev, prev_irev;
    COMPLEX_FLOAT y;
    const int n = fftsize;

    // 初期化
    if (A != a) {
        memcpy((void*) a, (void*) A, n * sizeof(COMPLEX_FLOAT));
    }
    COMPLEX_FLOAT *w = (COMPLEX_FLOAT *)malloc(n * sizeof(COMPLEX_FLOAT));
    if (w == NULL) {
        harkio_Log_printf(harkio_Log_Error, "malloc error");
        exit(EXIT_FAILURE);
    }

    for (i = 0; i < n; i++) {
        w[i] = cos(two_pi * i / n) - sin(two_pi * i / n) * _Complex_I;
    }


    // ビット反転並べ替え
    i = 0;
    for (j = 1; j < n - 1; j++) {
        for (k = n >> 1; k > (i ^= k); k >>= 1)
            ;
        if (j < i) {
            harktool_swapFloatComplex(&a[i], &a[j]); // std::swap(a[i], a[j]);
        }
    }

    // 主要計算部
    for (mh = 1; (m = mh << 1) <= n; mh = m) {
        irev = 0;
        for (i = 0; i < n; i += m) {
            prev_irev = irev;
            // w[prev_irev] = cos(two_pi * irev / n) - sin(two_pi * irev / n) * _Complex_I;
            for (k = n >> 2; k > (irev ^= k); k >>= 1)
                ;
            for (j = i; j < mh + i; j++) {
                k = j + mh;
                y = a[j] - a[k];
                a[j] = a[j] + a[k];
                a[k] = conjf(w[prev_irev]) * y;
            }
        }
    }

    // 終了処理
    free(w);
    w = NULL;
}

/*-----------------------------------------------------------------------------
*  実FFT 順計算       (float 実数 --> float 複素共役対称列)
*
*  float*  float_a    // 入力実数配列（time, 長さ n）
*  COMPLEX_FLOAT*  A  // 出力複素配列 (freq, 長さ n/2)
*  int fftsize        // FFT長 n
*  COMPLEX_FLOAT*  w  // 作業用配列（NULLの場合は関数内で確保する）
-----------------------------------------------------------------------------*/
// void harktool_FFTImpl_fRealFFT( float *float_a , COMPLEX_FLOAT  *A , int fft_size, COMPLEX_FLOAT* w );

/*-----------------------------------------------------------------------------
*  実FFT 逆計算       (float 実数 <-- float 複素共役対称列)
*
*  COMPLEX_FLOAT*  A  // 入力複素配列 (freq, 長さ n/2+1)
*  float*  float_a    // 出力実数配列（time, 長さ n）
*  int fftsize        // FFT長 n  ( n > 4 )
*  COMPLEX_FLOAT*  w  // 作業用配列（NULLの場合は関数内で確保する）
-----------------------------------------------------------------------------*/
void harktool_FFTImpl_fRealFFTInv(COMPLEX_FLOAT *A, float *float_a, int fft_size)
{

    COMPLEX_FLOAT y, z, w_k_real, w_k_imag;
    const int n = fft_size;
    const int n_2 = n / 2;
    const int n_4 = n / 4;


    COMPLEX_FLOAT *tmp_A; // 半サイズのFFT作業用

    tmp_A = (COMPLEX_FLOAT *)malloc((n_2) * sizeof(COMPLEX_FLOAT));
    if (tmp_A == NULL) {
        harkio_Log_printf(harkio_Log_Error, "malloc error");
        exit(EXIT_FAILURE);
    }


    // 変数変換
    tmp_A[0] = (crealf(A[0]) + crealf(A[n_2])) + (crealf(A[0]) - crealf(A[n_2])) * _Complex_I;
    tmp_A[0] = 0.5f * tmp_A[0]; // A[n_2]はここでのみ使う
    tmp_A[n_4] = conjf(A[n_4]);

    for (int k = 1; k < n_4; k++) {
        w_k_real = cosf(two_pi * k / n);
        w_k_imag = -sinf(two_pi * k / n);

        z = (1.0 + w_k_imag) + w_k_real * _Complex_I; // 1 + i*conj(Wk)
        y = 0.5 * z * (A[k] - conjf(A[n_2 - k]));

        tmp_A[k] = conjf(A[n_2 - k]) + y;
        tmp_A[n_2 - k] = conjf(A[k] - y);
    }

    // 長さ半分の複素IFFTを実行
    harktool_FFTImpl_fFFTInv(tmp_A, tmp_A, n_2);

    // 変数変換
    for (int i = 0; i < n_2; i++) {
        float_a[i * 2] = crealf(tmp_A[i]) * 2;
        float_a[i * 2 + 1] = cimagf(tmp_A[i]) * 2;
    }
}

/*-----------------------------------------------------------------------------
*  FFT 計算       (double complex --> double complex)
*
*  COMPLEX_DOUBLE*  a  // 入力列
*  COMPLEX_DOUBLE*  A  // 出力・計算用配列（in-place計算）
*  int fftsize         // FFT長
*  COMPLEX_DOUBLE*  w  // 作業用配列（NULLの場合は関数内で確保する）
-----------------------------------------------------------------------------*/
void harktool_FFTImpl_dFFT(COMPLEX_DOUBLE *a, COMPLEX_DOUBLE *A, int fftsize)
{
    int m, mh, i, j, k, irev, prev_irev;
    COMPLEX_DOUBLE y;
    const int n = fftsize;

    // 初期化
    if (a != A) {
        memcpy((void*) A, (void*) a, n * sizeof(COMPLEX_DOUBLE));
    }
    COMPLEX_DOUBLE *w = (COMPLEX_DOUBLE *)malloc(n * sizeof(COMPLEX_DOUBLE));
    if (w == NULL) {
        harkio_Log_printf(harkio_Log_Error, "malloc error");
        exit(EXIT_FAILURE);
    }

    for (i = 0; i < n; i++) {
        w[i] = cos(two_pi * i / n) - sin(two_pi * i / n) * _Complex_I;
    }

    // ビット反転並べ替え
    i = 0;
    for (j = 1; j < n - 1; j++) {
        for (k = n >> 1; k > (i ^= k); k >>= 1)
            ;
        if (j < i) {
            harktool_swapDoubleComplex(&A[i], &A[j]);
        }
    }

    // 主要計算部
    for (mh = 1; (m = mh << 1) <= n; mh = m) {
        irev = 0;
        for (i = 0; i < n; i += m) {
            prev_irev = irev;
            // w[prev_irev] = cos(two_pi * irev / n) - sin(two_pi * irev / n) * _Complex_I;
            for (k = n >> 2; k > (irev ^= k); k >>= 1)
                ;
            for (j = i; j < mh + i; j++) {
                k = j + mh;
                y = A[j] - A[k];
                A[j] = A[j] + A[k];
                A[k] = w[prev_irev] * y;
            }
        }
    }

    // 終了処理
    free(w);
    w = NULL;
}

/*-----------------------------------------------------------------------------
*  FFT 逆計算       (double complex <-- double complex)
*
*  COMPLEX_DOUBLE*  A  // 入力列
*  COMPLEX_DOUBLE*  a  // 出力・計算用配列（in-place計算）
*  int fftsize         // FFT長
*  COMPLEX_DOUBLE*  w  // 作業用配列（NULLの場合は関数内で確保する）
-----------------------------------------------------------------------------*/
void harktool_FFTImpl_dFFTInv(COMPLEX_DOUBLE *A, COMPLEX_DOUBLE *a, int fftsize)
{


    int m, mh, i, j, k, irev, prev_irev;
    // int len;/
    COMPLEX_DOUBLE y;

    const int n = fftsize;

    // 初期化
    if (A != a) {
        memcpy((void*) a, (void*) A, n * sizeof(COMPLEX_DOUBLE));
    }
    COMPLEX_DOUBLE *w = (COMPLEX_DOUBLE *)malloc(n * sizeof(COMPLEX_DOUBLE));
    if (w == NULL) {
        harkio_Log_printf(harkio_Log_Error, "malloc error");
        exit(EXIT_FAILURE);
    }

    for (i = 0; i < n; i++) {
        w[i] = cos(two_pi * i / n) - sin(two_pi * i / n) * _Complex_I;
    }


    // ビット反転並べ替え
    i = 0;
    for (j = 1; j < n - 1; j++) {
        for (k = n >> 1; k > (i ^= k); k >>= 1)
            ;
        if (j < i) {
            harktool_swapDoubleComplex(&a[i], &a[j]); // std::swap(a[i], a[j]);
        }
    }


    // 主要計算部
    for (mh = 1; (m = mh << 1) <= n; mh = m) {
        irev = 0;
        for (i = 0; i < n; i += m) {
            prev_irev = irev;
            // w[prev_irev] = cos(two_pi * irev / n) - sin(two_pi * irev / n) * _Complex_I;
            for (k = n >> 2; k > (irev ^= k); k >>= 1)
                ;
            for (j = i; j < mh + i; j++) {
                k = j + mh;
                y = a[j] - a[k];
                a[j] = a[j] + a[k];
                a[k] = conj(w[prev_irev]) * y;
            }
        }
    }

    //終了処理
    free(w);
    w = NULL;
}

/*-----------------------------------------------------------------------------
*  実FFT 順計算       (double 実数 --> double 複素共役対称列)
*
*  double*  double_a    // 入力実数配列（time, 長さ n）
*  COMPLEX_DOUBLE*  A  // 出力複素配列 (freq, 長さ n/2)
*  int fftsize         // FFT長 n
*  COMPLEX_DOUBLE*  w  // 作業用配列（NULLの場合は関数内で確保する）
-----------------------------------------------------------------------------*/
// void harktool_FFTImpl_dRealFFT( double *double_a , COMPLEX_DOUBLE  *A , int fft_size, COMPLEX_DOUBLE* w );

/*-----------------------------------------------------------------------------
*  実FFT 逆計算       (double 実数 <-- double 複素共役対称列)
*
*  COMPLEX_DOUBLE*  A  // 入力複素配列 (freq, 長さ n/2+1)
*  double*  double_a   // 出力実数配列（time, 長さ n）
*  int fftsize         // FFT長 n  ( n > 4 )
*  COMPLEX_DOUBLE*  w  // 作業用配列（NULLの場合は関数内で確保する）
-----------------------------------------------------------------------------*/
void harktool_FFTImpl_dRealFFTInv(COMPLEX_DOUBLE *A, double *double_a, int fft_size)
{


    COMPLEX_DOUBLE y, z, w_k_real, w_k_imag;


    const int n = fft_size;
    const int n_2 = n / 2;
    const int n_4 = n / 4;


    COMPLEX_DOUBLE *tmp_A; // 半サイズのFFT作業用

    tmp_A = (COMPLEX_DOUBLE *)malloc((n_2) * sizeof(COMPLEX_DOUBLE));
    if (tmp_A == NULL) {
        harkio_Log_printf(harkio_Log_Error, "malloc error");
        exit(EXIT_FAILURE);
    }


    // 変数変換
    tmp_A[0] = (creal(A[0]) + creal(A[n_2])) + (creal(A[0]) - creal(A[n_2])) * _Complex_I;
    tmp_A[0] = 0.5f * tmp_A[0]; // A[n_2]はここでのみ使う
    tmp_A[n_4] = conj(A[n_4]);

    for (int k = 1; k < n_4; k++) {
        w_k_real = cos(two_pi * k / n);
        w_k_imag = -sin(two_pi * k / n);

        z = (1.0 + w_k_imag) + w_k_real * _Complex_I; // 1 + i*conj(Wk)
        y = 0.5 * z * (A[k] - conj(A[n_2 - k]));

        tmp_A[k] = conj(A[n_2 - k]) + y;
        tmp_A[n_2 - k] = conj(A[k] - y);
    }

    // 長さ半分の複素IFFTを実行
    harktool_FFTImpl_dFFTInv(tmp_A, tmp_A, n_2);

    // 変数変換
    for (int i = 0; i < n_2; i++) {
        double_a[i * 2] = creal(tmp_A[i]) * 2;
        double_a[i * 2 + 1] = cimag(tmp_A[i]) * 2;
    }
}


/*----------------------------------------------------------------*
* float complex型　swap
*
-----------------------------------------------------------------*/
void harktool_swapFloatComplex(float _Complex  *a , float _Complex  *b)
{
    float _Complex tmp = *a;
    *a = *b;
    *b = tmp;
}



/*----------------------------------------------------------------*
* double complex型　swap
*
-----------------------------------------------------------------*/
void harktool_swapDoubleComplex(double _Complex  *a , double _Complex  *b)
{
    double _Complex tmp = *a;
    *a = *b;
    *b = tmp;
}
