#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <libharkio3.h>

#define FMTPATH_LEN 1024
#define TYPENAME_LEN 24

typedef enum { circle , grid , ball , cylinder, unknown } CoordModel_t;



//global
typedef struct {
    float min      ;
    float max      ;
    float interval ;
    int repnum     ;
} arg_t;



typedef struct {
    char modelname[TYPENAME_LEN];
    char posestype[TYPENAME_LEN];
    char output[FMTPATH_LEN];
    int  id           ;
    arg_t azimuth     ;
    arg_t elevation   ;
    arg_t height      ;
    arg_t radius      ;
    arg_t x_axis      ;
    arg_t y_axis      ;
    arg_t z_axis      ;
    char  path[FMTPATH_LEN];
    harkio_Positions *poses;
    CoordModel_t coordmodel;
} genposparam_t;
// global_t  g;



// values for formatting path str (do nothing if null)
typedef struct {
    float* pazimuth;
    float* pelevation;
    float* pheight;
    float* pradius;
    float* px_axis;
    float* py_axis;
    float* pz_axis;
} fmtval_t;


//Prototype
void initParameters(genposparam_t* m);

void getArguments(int argc , char** argv, genposparam_t* m);
int setValueStr(int* pcnt, char** argv, char *optstr, char* dst);
int setValueInt(int* pcnt, char** argv, char *optstr, int* idst);
int setValueFlt(int* pcnt, char** argv, char *optstr, float* fdst);

void checkArguments(genposparam_t* m);

void makePositionsFromModel(genposparam_t* m);   //
harkio_Position* polarToCartesian(harkio_Position* pos);
harkio_Position* cylindricalToCartesian(harkio_Position* pos);
void makeFilepath(const char *input, fmtval_t* fvals, char *output);
int validateFormatStr(const char* fmt);

void help(void);
void helpForModel(const char* coordstr);









int main(int argc, char **argv)
{

    harkio_XML *xml;
    FILE *fp;

    if (argc <= 1) {
        harkio_Log_printf(harkio_Log_Error , "argument error!!");
        help();
        exit(EXIT_FAILURE);
    } else if (argc == 2) {
        help();
        if (strcmp(argv[1] , "--help") == 0) {
        	exit(EXIT_SUCCESS);
        } else {
        	exit(EXIT_FAILURE);
        }
    } else if (strcmp(argv[2] , "--help") == 0) { //argc > 2
        helpForModel(argv[1]);
        exit(EXIT_SUCCESS);
    }


    genposparam_t prm;
    // init
    initParameters(&prm);


    //get argument
    getArguments(argc , argv, &prm);

    checkArguments(&prm);


    if ((fp = fopen(prm.output , "w")) == NULL) {
        harkio_Log_FileOpenFailed(prm.output);
        exit(EXIT_FAILURE);
    }


    prm.poses = harkio_Positions_new(prm.posestype , 0, 0, NULL);

    if (prm.poses == NULL) {
        harkio_Log_printf(harkio_Log_Error,"can not harkio_Positions_new\n") ;
        exit(EXIT_FAILURE);
    }

    makePositionsFromModel(&prm);

    // make new XML object and write
    xml = harkio_XML_newFromData(prm.poses, NULL, NULL);

    harkio_XML_write(xml, fp);

    harkio_XML_delete(&xml);

    fclose(fp);

    harkio_Log_printf(harkio_Log_Debug, "XML file created");


    exit(EXIT_SUCCESS);


}//mainend





void makePositionsFromModel(genposparam_t* m)
{
    CoordModel_t coordmodel = m->coordmodel;

    harkio_Position *pos ;
    char  msg[MSG_LEN];

    float radius;
    float elevation;
    float azimuth;
    float height;
    float x, y, z;
    int irad;
    int ielev;
    int iazim;
    int ihei;
    int ix, iy, iz;
    //init
    int id = m->id;
    char  formatted_path[FMTPATH_LEN];

    if (coordmodel == circle || coordmodel == ball) {
        // --------------------------------------------------------//
        //  circle
        //     方向角（最小）
        //     方向角（最大）
        //     方向角（間隔）
        //     高さ  （最小）
        //     高さ  （最大）
        //     高さ  （間隔）
        //     半径（最小）
        //     半径（最大）
        //     半径（間隔）
        // --------------------------------------------------------

        //半径(radius)
        for (irad=0; irad < m->radius.repnum; irad++) {
            radius = m->radius.min + irad * m->radius.interval;

            //高さ方向(elevation)
            for (ielev=0; ielev < m->elevation.repnum; ielev++) {
                elevation = m->elevation.min + ielev * m->elevation.interval;

                //方向角(azimuth)
                for (iazim=0; iazim < m->azimuth.repnum; iazim++) {
                    azimuth = m->azimuth.min + iazim * m->azimuth.interval;

                    sprintf(msg, "[%d] azimuth[%f] elevation[%f] radius[%f]\n",id,azimuth,elevation,radius);
                    harkio_Log_printf(harkio_Log_Debug, msg);

                    float coord[3] = {azimuth, elevation, radius};
                    fmtval_t fvals = {NULL, NULL, NULL, NULL, NULL, NULL, NULL};
                    fvals.pradius = &radius;
                    fvals.pelevation = &elevation;
                    fvals.pazimuth = &azimuth;
                    makeFilepath(m->path, &fvals, formatted_path);
                    pos = harkio_Position_new(id , Cartesian , coord , formatted_path);

                    if (pos == NULL) {
                        harkio_Log_printf(harkio_Log_Error,"can not harkio_Position_new\n") ;
                        exit(EXIT_FAILURE);
                    }

                    pos = polarToCartesian(pos); //極座標を直交座標 に変換する
                    harkio_Positions_append(m->poses , pos);

                    if (id != -1) {
                        id++;
                    }
                }
            }
        }
    } else if (coordmodel == grid) {
        // --------------------------------------------------------//
        // grid
        //     X方向（最小）
        //     X方向（最大）
        //     X方向（間隔）
        //     y方向（最小）
        //     y方向（最大）
        //     y方向（間隔）
        // --------------------------------------------------------

        for (iz=0; iz < m->z_axis.repnum; iz++) {
            z = m->z_axis.min + iz * m->z_axis.interval;

            for (ix=0; ix < m->x_axis.repnum; ix++) {
                x = m->x_axis.min + ix * m->x_axis.interval;

                for (iy=0; iy < m->y_axis.repnum; iy++) {
                    y = m->y_axis.min + iy * m->y_axis.interval;

                    sprintf(msg, "[%d] = (%f, %f, %f)", id, x, y, z);
                    harkio_Log_printf(harkio_Log_Debug, msg);

                    float coord[3] = {x, y, z};
                    fmtval_t fvals = {NULL, NULL, NULL, NULL, NULL, NULL, NULL};
                    fvals.px_axis = &x;
                    fvals.py_axis = &y;
                    fvals.pz_axis = &z;
                    makeFilepath(m->path, &fvals, formatted_path);
                    pos = harkio_Position_new(id , Cartesian , coord , formatted_path);

                    if (pos == NULL) {
                        harkio_Log_printf(harkio_Log_Error,"can not harkio_Position_new\n") ;
                        exit(EXIT_FAILURE);
                    }

                    harkio_Positions_append(m->poses , pos);

                    if (id != -1) {
                        id++;
                    }
                }//y
            }//x
        }//z
    } else if ((coordmodel == cylinder)) {
        //--------------------------------------------------------//
        // cylinder
        //    方向角（最小）
        //    方向角（最大）
        //    方向角（間隔）
        //    高さ  （最小）
        //    高さ  （最大）
        //    高さ  （間隔）
        //    半径  （最小）
        //    半径  （最大）
        //    半径  （間隔）
        //--------------------------------------------------------//

        //半径(radius)
        for (irad=0; irad < m->radius.repnum; irad++) {
            radius = m->radius.min + irad * m->radius.interval;

            //高さ方向(height)
            for (ihei=0; ihei < m->height.repnum; ihei++) {
                height = m->height.min + ihei * m->height.interval;

                //横方向(azimuth)
                for (iazim=0; iazim < m->azimuth.repnum; iazim++) {
                    azimuth = m->azimuth.min + iazim * m->azimuth.interval;

                    sprintf(msg, "[%d] azimuth[%f] height[%f] radius[%f]\n",id,azimuth,height,radius);
                    harkio_Log_printf(harkio_Log_Debug, msg);

                    float coord[3] = {azimuth, radius, height};
                    fmtval_t fvals = {NULL, NULL, NULL, NULL, NULL, NULL, NULL};
                    fvals.pradius = &radius;
                    fvals.pheight = &height;
                    fvals.pazimuth = &azimuth;
                    makeFilepath(m->path, &fvals, formatted_path);
                    pos = harkio_Position_new(id , Cartesian , coord , formatted_path);

                    if (pos == NULL) {
                        harkio_Log_printf(harkio_Log_Error,"can not harkio_Position_new\n") ;
                        exit(EXIT_FAILURE);
                    }

                    cylindricalToCartesian(pos);
                    harkio_Positions_append(m->poses , pos);

                    if (id != -1) {
                        id++;
                    }
                }
            }
        }
    } else {
        harkio_Log_printf(harkio_Log_Error,"Unknown coordinate model\n") ;
        exit(EXIT_FAILURE);
    }
}




harkio_Position* polarToCartesian(harkio_Position* pos)
{
    float coord[3];
    if (pos != NULL) {
        //極座標を直交座標 に変換する
        harkio_Position_Convert(pos , coord , Cartesian);

        //positionにセット
        pos->coord[0] = coord[0];
        pos->coord[1] = coord[1];
        pos->coord[2] = coord[2];
    }
    return pos;
}


harkio_Position* cylindricalToCartesian(harkio_Position* pos)
{
    if (pos != NULL) {
        float azimuth = pos->coord[0];
        float radius = pos->coord[1];
        float height = pos->coord[2];
        float azimuth_radian = harkio_Position_convToRadian(azimuth);

        pos->coord[0] = radius * cosf(azimuth_radian);
        pos->coord[1] = radius * sinf(azimuth_radian);
        pos->coord[2] = height;
    }
    return pos;
}




//--------------------------------------------------------//
//Make path

//--------------------------------------------------------//
void makeFilepath(const char *input, fmtval_t* fvals, char *output)
{

    strcpy(output, "");
    if (strlen(input) == 0) {
        return;
    }

    char workstr[FMTPATH_LEN];
    strcpy(workstr, input);
    float fvalue;

    char *ss = workstr;         // ss: starting point
    char *pp, *qq, *rr;         // pp,qq,rr are pointers for '{', ':', '}' respectively
    char formatted_val[FMTPATH_LEN];

    //                                                  // "FILE_{AZIMUTH:%03f}_{RADIUS:%03f}.wav$" (example)
    while (ss != NULL && *ss != '\0') {                 // "s    |       |    |                   " ("$" means '\0')
        //                                              // "|    |       |    |                   "
        pp = strchr(ss, '{');                           // "s    p       |    |                  $"
        if (pp == NULL) {                               // "|    |       |    |                   "
            strcat(output, ss);                         // (if no '{' found, copy remaining str all then finish)
            break;                                      // "|    |       |    |                   "
        }                                               // "|    |       |    |                   "
        qq = strchr(pp, ':');                           // "s    p       q    |                  $"
        rr = strchr(pp, '}');                           // "s    p       q    r                  $"
        //                                              // "|    |       |    |                   "
        if (qq == NULL || rr == NULL || rr < qq) {
            harkio_Log_printf(harkio_Log_Error, "path string error: invalid after '{'");
            harkio_Log_printf(harkio_Log_Error, pp);
            exit(EXIT_FAILURE);                         // "|    |       |    |                   "
        }                                               // "|    |       |    |                   "
        // "s    p       q    r                   "
        *pp = *qq = *rr = '\0';                         // "FILE_$AZIMUTH$%03f$_{RADIUS:%03f}.wav$" ("$" means '\0')
        pp++;                                           // "|    |       |    |                   "
        qq++;                                           // "|    |       |    |                   "
        rr++;                                           // "s    $p      $q   $r                 $"

        if (validateFormatStr(qq) == EXIT_FAILURE) {
            harkio_Log_printf(harkio_Log_Error, "path string error: invalid format string");
            harkio_Log_printf(harkio_Log_Error, qq);
            exit(EXIT_FAILURE);
        }

        // '{'前までの文字列を追加                      // "FILE_$AZIMUTH$%03f$_{RADIUS:%03f}.wav$"
        strcat(output, ss);                             // "s    |       |    |                   "
        //                                              // (s=FILE_$)
        ss = rr;                                        // "|    |p      $q   $s                 $"
        //                                              // (p=AZIMUTH$,q=%03f$,
        //                                              // s=_{RADIUS:%03f}.wav$) (for next loop)

        if (strcmp(pp ,"AZIMUTH") == 0 && fvals->pazimuth) {
            fvalue = *(fvals->pazimuth);
            sprintf(formatted_val , qq , fvalue);
        } else if (strcmp(pp  ,"RADIUS") == 0 && fvals->pradius) {
            fvalue = *(fvals->pradius);
            sprintf(formatted_val , qq , fvalue);
        } else if (strcmp(pp  ,"ELEVATION") == 0 && fvals->pelevation) {
            fvalue = *(fvals->pelevation);
            sprintf(formatted_val , qq , fvalue);
        } else if (strcmp(pp ,"HEIGHT") == 0 && fvals->pheight) {
            fvalue = (int)(roundf(*(fvals->pheight) * 1000));
            sprintf(formatted_val , qq , fvalue);
        } else if (strcmp(pp  , "X_AXIS") == 0 && fvals->px_axis) {
            fvalue = (int)(roundf(*(fvals->px_axis) * 1000));
            sprintf(formatted_val , qq , fvalue);
        } else if (strcmp(pp  , "Y_AXIS") == 0 && fvals->py_axis) {
            fvalue = (int)(roundf(*(fvals->py_axis) * 1000));
            sprintf(formatted_val , qq , fvalue);
        } else if (strcmp(pp  , "Z_AXIS") == 0 && fvals->pz_axis) {
            fvalue = (int)(roundf(*(fvals->pz_axis) * 1000));
            sprintf(formatted_val , qq , fvalue);
        } else {
            sprintf(formatted_val, "{%s:%s}", pp, qq);
        }
        strcat(output, formatted_val);
    }
}


int validateFormatStr(const char* fmt)
{
    // 始端が[%] 終端が[feEgG] その他は数字か特定の記号であれば、okとする
    int valid_fmt = EXIT_FAILURE;
    int ii = 0;
    if (fmt[ii] == '%') {
        ii = 1;
        for (; ii<strlen(fmt) - 1; ii++) {                       // strlen(fmt) >= 1 == strlen('%')
            if (strchr("0123456789-. +#", fmt[ii]) == NULL) {    // fmt[ii] is digit or in {-. +#}      {
                break;
            }
        }

        if (fmt[ii] != '\0' && strchr("feEgG", fmt[ii]) != NULL) {
            if (fmt[ii + 1] == '\0') {
                valid_fmt = EXIT_SUCCESS;
            }
        }
    }
    return valid_fmt;


}






void initParameters(genposparam_t *m)
{

    m->id                 = 0;
    m->azimuth.min        = 0;
    m->elevation.min      = 0;
    m->height.min         = 0;
    m->radius.min         = 0;
    m->x_axis.min         = 0;
    m->y_axis.min         = 0;
    m->z_axis.min         = 0;

    m->azimuth.max        = 0;
    m->elevation.max      = 0;
    m->height.max         = 0;
    m->radius.max         = 0;
    m->x_axis.max         = 0;
    m->y_axis.max         = 0;
    m->z_axis.max         = 0;

    m->azimuth.interval   = 0;
    m->elevation.interval = 0;
    m->height.interval    = 0;
    m->radius.interval    = 0;
    m->x_axis.interval    = 0;
    m->y_axis.interval    = 0;
    m->z_axis.interval    = 0;

    memset(m->modelname, '\0' , sizeof(m->modelname));
    memset(m->posestype, '\0' , sizeof(m->posestype));
    memset(m->output   , '\0' , sizeof(m->output));

}

int setValueFlt(int* pcnt, char** argv, char *optstr, float* fdst)
{
    // *pcnt目の引数がoptstrに等しければ、値をfdstに代入する
    char *endptr;

    if (strcmp(argv[*pcnt], optstr) == 0) {
        ++(*pcnt);
        if (argv[*pcnt] == NULL) {
            harkio_Log_printf(harkio_Log_Error, "option parse error");
            exit(EXIT_FAILURE);
        }
        *fdst = strtof(argv[*pcnt], &endptr);
        printf("%-24s[%f]\n", optstr, *fdst);

        ++(*pcnt);
        return 1; //true
    } else {
        return 0; //false
    }
}

int setValueInt(int* pcnt, char** argv, char *optstr, int* idst)
{
    // *pcnt目の引数がoptstrに等しければ、値をidstに代入する
    if (strcmp(argv[*pcnt], optstr) == 0) {
        ++(*pcnt);
        if (argv[*pcnt] == NULL) {
            harkio_Log_printf(harkio_Log_Error, "option parse error");
            exit(EXIT_FAILURE);
        }
        *idst = atoi(argv[*pcnt]);
        printf("%-24s[%d]\n", optstr, *idst);

        ++(*pcnt);
        return 1; //true
    } else {
        return 0; //false
    }
}

int setValueStr(int* pcnt, char** argv, char *optstr, char* dst)
{
    // *pcnt目の引数がoptstrに等しければ、値をdstに書く
    if (strcmp(argv[*pcnt], optstr) == 0) {
        ++(*pcnt);
        if (argv[*pcnt] == NULL) {
            harkio_Log_printf(harkio_Log_Error, "option parse error");
            exit(EXIT_FAILURE);
        }
        strcpy(dst, argv[*pcnt]);
        printf("%-24s[%s]\n", optstr, dst);

        ++(*pcnt);
        return 1; //true
    } else {
        return 0; //false
    }
}


void getArguments(int argc , char** argv, genposparam_t* m)
{
    int count;
    float elevation_tmp;
    CoordModel_t coordmodel;

    // printf("%s \n",argv[0]);

    strcpy(m->modelname , argv[1]);

    if (strcmp(m->modelname, "circle") == 0) {
        m->coordmodel = circle;
    } else if (strcmp(m->modelname, "grid") == 0) {
        m->coordmodel = grid;
    } else if (strcmp(m->modelname, "ball") == 0) {
        m->coordmodel = ball;
    } else if (strcmp(m->modelname, "cylinder") == 0) {
        m->coordmodel = cylinder;
    } else {
        harkio_Log_printf(harkio_Log_Error,"coordinate model error") ;
        exit(EXIT_FAILURE);
    }
    coordmodel = m->coordmodel;

    printf("[ %s : ] \n", m->modelname);


    count = 2;

    while (count < argc) {
        //coord positions type
        if (setValueStr(&count, argv, "--type"              , m->posestype)) continue;
        //output
        if (setValueStr(&count, argv, "--output"            , m->output)) continue;
        //id
        if (setValueInt(&count, argv, "--id"                ,&m->id)) continue;
        //path
        if (setValueStr(&count, argv, "--path"              , m->path)) continue;

        if (coordmodel == ball || coordmodel == circle || coordmodel == cylinder) {
            //方向角（最小）
            if (setValueFlt(&count, argv, "--min-azimuth"       ,&m->azimuth.min)) continue;
            //方向角（最大）
            if (setValueFlt(&count, argv, "--max-azimuth"       ,&m->azimuth.max)) continue;
            //方向角（間隔）
            if (setValueFlt(&count, argv, "--interval-azimuth"  ,&m->azimuth.interval)) continue;
            //半径（最小）
            if (setValueFlt(&count, argv, "--min-radius"        ,&m->radius.min)) continue;
            //半径（最大）
            if (setValueFlt(&count, argv, "--max-radius"        ,&m->radius.max)) continue;
            //半径（間隔）
            if (setValueFlt(&count, argv, "--interval-radius"   ,&m->radius.interval)) continue;
        }
        if (coordmodel == ball || coordmodel == circle) { // circle for compat
            //仰角  （最小）
            if (setValueFlt(&count, argv, "--min-elevation"     ,&m->elevation.min)) continue;
            //仰角  （最大）
            if (setValueFlt(&count, argv, "--max-elevation"     ,&m->elevation.max)) continue;
            //仰角  （間隔）
            if (setValueFlt(&count, argv, "--interval-elevation",&m->elevation.interval)) continue;
        }
        if (coordmodel == cylinder) {
            //高さ  （最小）
            if (setValueFlt(&count, argv, "--min-height"        ,&m->height.min)) continue;
            //高さ  （最大）
            if (setValueFlt(&count, argv, "--max-height"        ,&m->height.max)) continue;
            //高さ  （間隔）
            if (setValueFlt(&count, argv, "--interval-height"   ,&m->height.interval)) continue;
        }
        if (coordmodel == grid) {
            //x方向（最小）
            if (setValueFlt(&count, argv, "--min-x_axis"        ,&m->x_axis.min)) continue;
            //x方向（最大）
            if (setValueFlt(&count, argv, "--max-x_axis"        ,&m->x_axis.max)) continue;
            //x方向（間隔）
            if (setValueFlt(&count, argv, "--interval-x_axis"   ,&m->x_axis.interval)) continue;
            //y方向（最小）
            if (setValueFlt(&count, argv, "--min-y_axis"        ,&m->y_axis.min)) continue;
            //y方向（最大）
            if (setValueFlt(&count, argv, "--max-y_axis"        ,&m->y_axis.max)) continue;
            //y方向（間隔）
            if (setValueFlt(&count, argv, "--interval-y_axis"   ,&m->y_axis.interval)) continue;
            //z方向（最小）
            if (setValueFlt(&count, argv, "--min-z_axis"        ,&m->z_axis.min)) continue;
            //z方向（最大）
            if (setValueFlt(&count, argv, "--max-z_axis"        ,&m->z_axis.max)) continue;
            //z方向（間隔）
            if (setValueFlt(&count, argv, "--interval-z_axis"   ,&m->z_axis.interval)) continue;
        }
        if (coordmodel == circle) {
            //高さ（固定）
            if (setValueFlt(&count, argv, "--elevation"            ,&elevation_tmp)) {
                m->elevation.interval = 0;
                m->elevation.min = elevation_tmp;
                m->elevation.max = elevation_tmp;
                continue;
            }
        }

        //none of them
        char msg[MSG_LEN];
        sprintf(msg, "Invalid argument for %s: %s", m->modelname, argv[count]);
        harkio_Log_printf(harkio_Log_Error, msg) ;
        // help();
        exit(EXIT_FAILURE);
    }

    // set default values
    m->height.interval    = (m->height.interval    != 0) ? m->height.interval    : 0.01;
    m->radius.interval    = (m->radius.interval    != 0) ? m->radius.interval    : 1;
    m->azimuth.interval   = (m->azimuth.interval   != 0) ? m->azimuth.interval   : 5;
    m->elevation.interval = (m->elevation.interval != 0) ? m->elevation.interval : 16.70;
    m->x_axis.interval    = (m->x_axis.interval    != 0) ? m->x_axis.interval    : 1;
    m->y_axis.interval    = (m->y_axis.interval    != 0) ? m->y_axis.interval    : 1;
    m->z_axis.interval    = (m->z_axis.interval    != 0) ? m->z_axis.interval    : 1;

    if (m->azimuth.min == 0.f && m->azimuth.max == 360.f) {
        m->azimuth.max -= m->azimuth.interval;
    }

    // calc repeating numbers
    m->elevation.repnum = 1 + (int) floorf((m->elevation.max - m->elevation.min) / m->elevation.interval);
    m->radius.repnum    = 1 + (int) floorf((m->radius.max    - m->radius.min)    / m->radius.interval);
    m->azimuth.repnum   = 1 + (int) floorf((m->azimuth.max   - m->azimuth.min)   / m->azimuth.interval);
    m->height.repnum    = 1 + (int) floorf((m->height.max    - m->height.min)    / m->height.interval);
    m->x_axis.repnum    = 1 + (int) floorf((m->x_axis.max    - m->x_axis.min)    / m->x_axis.interval);
    m->y_axis.repnum    = 1 + (int) floorf((m->y_axis.max    - m->y_axis.min)    / m->y_axis.interval);
    m->z_axis.repnum    = 1 + (int) floorf((m->z_axis.max    - m->z_axis.min)    / m->z_axis.interval);

}





void checkArguments(genposparam_t* m)
{

    //
    if (strlen(m->modelname)  == 0) {
        harkio_Log_printf(harkio_Log_Error,"argument error [coord model name]\n") ;
        exit(EXIT_FAILURE);
    }

    //tsp/microphone
    if (strlen(m->posestype)    == 0) {
        harkio_Log_printf(harkio_Log_Error,"argument error [positions type]\n") ;
        exit(EXIT_FAILURE);
    }

    if (strlen(m->output)  == 0) {
        harkio_Log_printf(harkio_Log_Error,"argument error [no output]\n") ;
        exit(EXIT_FAILURE);
    }

    // coord
    const char *msg = NULL;
    CoordModel_t coordmodel = m->coordmodel;

    if (coordmodel == circle || coordmodel == ball || coordmodel == cylinder) {
        //距離
        if (!msg && !(m->radius.min > 0)) msg = "min-radius > 0";
        if (!msg && !(m->radius.max > 0)) msg = "max-radius > 0";
        if (!msg && !(m->radius.min <= m->radius.max)) msg = "min-radius <= max-radius";


        //仰角または高さ
        if (coordmodel == circle || coordmodel == ball) {
            if (!msg && !(m->elevation.min <= m->elevation.max)) msg = "min-elevation <= max-elevation";
        } else if (coordmodel == cylinder) {
            if (!msg && !(m->height.min <= m->height.max)) msg = "min-height <= max-height";
        }


        //方向角
        if (!msg && !(m->azimuth.min < m->azimuth.max)) msg = "min-azimuth < max-azimuth";
        if (!msg && !(m->azimuth.interval > 0)) msg = "interval-azimuth > 0";
    } else if (coordmodel == grid) {
        if (!msg && !(m->x_axis.min < m->x_axis.max)) msg = "min-x_axis < max-x_axis";
        if (!msg && !(m->x_axis.interval > 0)) msg = "interval-x_axis > 0";
        if (!msg && !(m->y_axis.min < m->y_axis.max)) msg = "min-y_axis < max-y_axis";
        if (!msg && !(m->y_axis.interval > 0)) msg = "interval-x_axis > 0";
        if (!msg && !(m->z_axis.min <= m->z_axis.max)) msg = "min-z_axis < max-z_axis";
    } else {
        harkio_Log_printf(harkio_Log_Error,"coordinate model error") ;
        exit(EXIT_FAILURE);
    }

    if (msg) {
        char msgs[MSG_LEN];
        sprintf(msgs, "must be [%s]", msg);
        harkio_Log_printf(harkio_Log_Error, msgs);
        exit(EXIT_FAILURE);
    }
    return;
}





/*----------------------------------------------------------------*/
//
//
/*----------------------------------------------------------------*/
void help(void)
{
    printf("                                                                            \n");
    printf("                                                                            \n");
    printf("[NAME:]                                                                     \n");
    printf("  harktoolcli-genposxml                                                     \n");
    printf("                                                                            \n");
    printf("[Description:]                                                              \n");
    printf("   create a position tag of xml file                                        \n");
    printf("                                                                            \n");
    printf("[Usage :]                                                                   \n");
    printf("   harktoolcli-genposxml [coordinate model] [options]                       \n");
    printf("                                                                            \n");
    printf("[Usage of each coordinate model :]                                          \n");
    printf("   harktoolcli-genposxml circle   --help                                    \n");
    printf("   harktoolcli-genposxml grid     --help                                    \n");
    printf("   harktoolcli-genposxml cylinder --help                                    \n");
    printf("   harktoolcli-genposxml ball     --help                                    \n");
    printf("                                                                            \n");
    printf("                                                                            \n");
}





/*----------------------------------------------------------------*/
//
//
/*----------------------------------------------------------------*/
void helpForModel(const char* coordstr)
{
    const char* pathvar[3];
    CoordModel_t coordmodel;

    if (strcmp(coordstr, "circle") == 0) {
        coordmodel = circle;
        pathvar[0] = "D{AZIMUTH:%03.1f}";
        pathvar[1] = "E{ELEVATION:%03.1f}";
        pathvar[2] = "R{RADIUS:%03.1f}";
    } else if (strcmp(coordstr, "grid") == 0) {
        coordmodel = grid;
        pathvar[0] = "X{X_AXIS:%03.1f}";
        pathvar[1] = "Y{Y_AXIS:%03.1f}";
        pathvar[2] = "Z{Z_AXIS:%03.1f}";
    } else if (strcmp(coordstr, "ball") == 0) {
        coordmodel = ball;
        pathvar[0] = "D{AZIMUTH:%03.1f}";
        pathvar[1] = "E{ELEVATION:%03.1f}";
        pathvar[2] = "R{RADIUS:%03.1f}";
    } else if (strcmp(coordstr, "cylinder") == 0) {
        coordmodel = cylinder;
        pathvar[0] = "D{AZIMUTH:%03.1f}";
        pathvar[1] = "H{HEIGHT:%03.1f}";
        pathvar[2] = "R{RADIUS:%03.1f}";
    } else {
        help();
        return;
    }

    printf("                                                                            \n");
    printf("                                                                            \n");
    printf("[coordinate model is %s:] \n", coordstr);
    printf("                                                                            \n");
    printf("  harktoolcli-genposxml %s [options] \n", coordstr);
    printf("                                                                            \n");
    printf("[options]                                                                   \n");
    printf("  --type               Type of positions                                    \n");
    printf("                        - tsp / impulse / noise / microphone / soundsource  \n");
    printf("  --output             Output filename                                      \n");
    printf("  --id                 Position id of start                                 \n");
    printf("                        - id == -1:  all set [-1]  / id >= 0 : increment    \n");
    printf("  --path               wav file path                                        \n");
    printf("                        - ex. filename_%s_%s_%s \n", pathvar[0], pathvar[1], pathvar[2]);

    if (coordmodel == circle || coordmodel == ball || coordmodel == cylinder) {
        printf("  --min-azimuth        Minimum  of azimuth                                  \n");
        printf("  --max-azimuth        Maximum  of azimuth                                  \n");
        printf("  --interval-azimuth   Interval of azimuth                                  \n");
        if (coordmodel == circle) {
            printf("  --elevation          Value of elevation                                   \n");
        }
        if (coordmodel == cylinder) {
            printf("  --min-height         Minimum  of height                                   \n");
            printf("  --max-height         Maximum  of height                                   \n");
            printf("  --interval-height    Interval of height                                   \n");
        } else if (coordmodel == ball) {
            printf("  --min-elevation      Minimum  of elevation                                \n");
            printf("  --max-elevation      Maximum  of elevation                                \n");
            printf("  --interval-elevation Interval of elevation                                \n");
        }
        printf("  --min-radius         Minimum  of radius                                   \n");
        printf("  --max-radius         Maximum  of radius                                   \n");
        printf("  --interval-radius    Interval of radius                                   \n");
    } else if (coordmodel == grid) {
        printf("  --min-x_axis         Minimum  of x_axis                                   \n");
        printf("  --max-x_axis         Maximum  of x_axis                                   \n");
        printf("  --interval-x_axis    Interval of x_axis                                   \n");
        printf("  --min-y_axis         Minimum  of y_axis                                   \n");
        printf("  --max-y_axis         Maximum  of y_axis                                   \n");
        printf("  --interval-y_axis    Interval of y_axis                                   \n");
        printf("  --min-z_axis         Minimum  of z_axis                                   \n");
        printf("  --max-z_axis         Maximum  of z_axis                                   \n");
        printf("  --interval-z_axis    Interval of z_axis                                   \n");
    }
    printf("                                                                            \n");
    printf("                                                                            \n");
}


